:- multifile pfd_rule/2.
% :- ensure_loaded( '../../src/pl' ).
:- ensure_loaded( 'vertical_tree' ).

% occurancies( [15890, 5338, 9657, 8359, 20783, 3357, 5738, 4328, 15569, 164, 1145, 9282, 4264, 13225, 10551, 3151, 241, 11957, 12422, 11065, 5107, 1594, 835, 727, 2832, 730] ).
% frequencies( [2270/25473, 5338/178311, 3219/59437, 8359/178311, 2969/25473, 1119/59437, 5738/178311, 4328/178311, 15569/178311, 164/178311, 1145/178311, 442/8491, 4264/178311, 13225/178311, 3517/59437, 3151/178311, 241/178311, 11957/178311, 12422/178311, 11065/178311, 5107/178311, 1594/178311, 835/178311, 727/178311, 944/59437, 730/178311] ).

/* 
unscrambled( [
 "administratively", "amicably", "amongst", "amplification", "attacher",
 "bitternut", "blacklist", "blends", "carbonization", "chattel", 
 "chaucer", "clothesline", "code", "cohered", "cult",
 "cycling", "eardrum", "ensuing", "euphoria", "everglade",
 "extremity", "felder", "game", "hartley", "hicks",
 "impelling", "jaycee", "lowliest", "measured", "mellowness",
 "mettlesome", "miscegenation", "mutant", "parts", "pictured",
 "rarity", "recruit", "robbery", "security", "shames",
 "speculations", "syntactic", "thinker", "trowels", "using",
 "vandalizing", "venison", "verbal", "writable", "xerography"]).

scrambled( [ 
	"tawcdcmvytvclope", "twcbtspe", "twfdgmv", "twrpcicbtvcfd", "tvvtbhoy",
	"scvvoydnv", "sptbupcmv", "spodam", "btysfdcqtvcfd", "bhtvvop",
	"bhtnboy", "bpfvhompcdo", "bfao", "bfhoyoa", "bnpv",
	"bebpcdg", "otyaynw", "odmncdg", "onrhfyct", "oloygptao",
	"ojvyowcve", "iopaoy", "gtwo", "htyvpoe", "hcbum",
	"cwroppcdg", "zteboo", "pfxpcomv", "wotmnyoa", "woppfxdomm",
	"wovvpomfwo", "wcmbogodtvcfd", "wnvtdv", "rtyvm", "rcbvnyoa",
	"ytycve", "yobyncv", "yfssoye", "mobnycve", "mhtwom",
	"mrobnptvcfdm", "medvtbvcb", "vhcduoy", "vyfxopm", "nmcdg",
	"ltdatpcqcdg", "lodcmfd", "loystp", "xycvtspo", "joyfgytrhe"
	] ).

*/

:- ensure_loaded( 'select_words' ). % /3.
:- ensure_loaded( 'lower_case_words' ). % /3.
% :- ensure_loaded( 'scramble_words' ). % /3.
:- ensure_loaded( 'count_letter_occurances_in_words' ). % /3.

:- ensure_loaded(library(lists)).
:-  ensure_loaded('std').

% from_file_read_term( dicts, dict(D,DictOrig,Freqs) ),
from_file_read_term( File, Term ) :-
	open( File, read, Stream ),
	from_stream_read_term( Stream, Term ).

from_stream_read_term( Stream, Term ) :-
	( read( Stream, Input ) -> 
		( Input = Term ->
			close( Stream )
			;
			from_stream_read_term( Stream, Term )
		)
		;
		close( Stream ), fail
	).

statistic_pair( System, Time ) :-
	pl( System ),
	( System = swi(_Some) -> 
		statistics(cputime,Time)
		;
		statistics(runtime,[RTime,_SLRT]),
		statistics(walltime,[WTime,_SLWT]),
		Time = RTime-WTime
	).

atom_codes_list( [], [] ).
atom_codes_list( [H|T], [HCs|TCs] ) :-
	atom_codes( H, HCs ),
	atom_codes_list( T, TCs ).

% :- ensure_loaded( library(lists) ). % nth1/4.
% :- ensure_loaded( library('meta/maplist') ). % maplist/3. assume it 

% simple_test( Pr ) :-
	% Var pin proximity([a,b,c],[5/12,1/4,1/3],1/2),
	% Pr is p(Var=b).
make_probs( Probs ) :-
	occurancies( Freqs ),
	maplist( rationals_integer_to, Freqs, Rats ),
	rationals_to_probabilities( Rats, Probs ).


export_words( Legend, Words ) :-
	write( Legend ), write( ': ' ),
	maplist( name, Writes, Words ),
	write( Writes ).
	% write( words(Writes) ).
	% write( words(Writes) ), nl.

word_codes_to_vars( [], _Codes, _Vars, [] ) :- 
	!.
word_codes_to_vars( [H|T], Codes, Vars, [HCode|TCode] ) :-
	!,
	word_codes_to_vars( H, Codes, Vars, HCode ),
	word_codes_to_vars( T, Codes, Vars, TCode ).
word_codes_to_vars( Code, Codes, Vars, Var ) :-
	nth1( Nth1, Codes, Code ),
	nth1( Nth1, Vars, Var ).

wordcodes_to_vars( [], _Codes, _Vars, [] ).
wordcodes_to_vars( [H|T], Codes, Vars, WordVars ) :-
	( nth1( Nth1, Codes, H, NxCodes ) -> 
		% new code, remove from pool
		nth1( Nth1, Vars, VarH, NxVars ),
		WordVars = [VarH|VarT]
		;
		% the Code has already been seen.
		WordVars = VarT,
		NxCodes = Codes,
		NxVars = Vars
	),
	!, % dont backtract
	wordcodes_to_vars( T, NxCodes, NxVars, VarT ).

% word_vars_to_codes( WordVs, Vars, Codes, WordCs ),
word_vars_to_codes( [], _Vars, _Codes, [] ).
word_vars_to_codes( [HV|TVs], Vars, Codes, [HC|TCs] ) :-
	nth1_vars( Vars, HV, 1, Nth1 ),
	nth1( Nth1, Codes, HC ),
	word_vars_to_codes( TVs, Vars, Codes, TCs ).

% nth1_vars( [], _V, Acc, Nth1 ) :-
nth1_vars( [H|T], V, Acc, Nth1 ) :-
	( H==V -> 
		Nth1 is Acc
		;
		NxAcc is Acc + 1,
		nth1_vars( T, V, NxAcc, Nth1 )
	).

new_wordcodes_to_guess( [], WordVls, _Vals, WordVls ).
new_wordcodes_to_guess( [Hvr|Tvrs], WordVrs, [Hvl|Tvls],  WordVls ) :-
	substitute( Hvr, WordVrs, Hvl, NxWordVrs ),
	new_wordcodes_to_guess( Tvrs, NxWordVrs, Tvls, WordVls ).

wordcodes_to_guess( [], _Codes, _Vars, _WordVs, _Vals, [] ).
wordcodes_to_guess( [H|T], Codes, Vars, WordVs, Vals, [HVl|TGuess] ) :-
	nth1( Nth1, Codes, H ),
	nth1( Nth1, Vars, HVr ),
	nth1( WVNth1, WordVs, HVr ),
	nth1( WVNth1, Vals, HVl ),
	wordcodes_to_guess( T, Codes, Vars, WordVs, Vals, TGuess ).

% :- ensure_loaded( '../../load_pfd' ).

test :- 
	scrambled( Scrambled ),
	export_words( 'Scrambled words', Scrambled ), nl,
	test( Scrambled, Solution ), 
	export_words( 'Original words', Scrambled ), nl,
	export_words( 'Read as ', Solution ).

main :- 
	on_pl( sicstus(_Sics), prolog_flag(argv,UnixArgs) ),
	on_pl( yap(_Yap), unix(argv(UnixArgs)) ),
	UnixArgs  = [Datom,Satom|_Empty],
	atom_chars( Datom, Dcs ),
	number_chars( D, Dcs ),
	atom_chars( Satom, Scs ),
	number_chars( S, Scs ),
	main_caesar( D, S ).

main_caesar( D, S ) :- 
	from_file_read_term( dicts, dict(D,DictOrig,Freqs) ),
	from_file_read_term( selections, selection(S,HowMany,DictWords) ),
	from_file_read_term( selections, encoded(S,HowMany,_Map,AllEncWords,Counts,Sum) ),
	atom_codes_list( DictWords, DictWCs ),
	% assert( unscrambled(DictWCs) ),
	bb_put( unscrambled, DictWCs ),
	four( AllEncWords, EncWords ),
	write( four(EncWords) ), nl,
	atom_codes_list( EncWords, EncWCs ),
	alphabet( AlphaBeto ),
	statistic_pair( System, PreTime ),
	write( entering(caesar(Counts)) ),
	demo( [caesar( EncWCs, Counts, Sum, AlphaBeto, Freqs, _Vars, SolutionCs )] ),
	maplist( name, Solution, SolutionCs ),
	statistic_pair( System, PostTime ),
	Time = (PreTime,PostTime),
	get_host_name_datime( Host, Date ),
	open( register_four, append, Stream ),
	write_term( Stream, res(Date,D,S,DictOrig,Host,System,Time,Solution), [quoted(true)] ),
	write(Stream,'.'), nl(Stream),
	close( Stream ),
	write( done ), nl.

four( [], [] ).
four( [W|Ws], FourWords ) :-
	atom_codes( W, WCs ),
	sort( WCs, Uniqcs ),
	( Uniqcs = [_,_,_,_|_] ->
		FourWords = [W|MoreFourWords]
		;
		write( discarding(W) ), nl,
		FourWords = MoreFourWords
	),
	four( Ws, MoreFourWords ).

get_host_name_datime( Host, Date ) :- 
	( pl(swi(_)) ->
		current_prolog_flag( arch, Arch ),
		get_time( Time ),
		convert_time(Time,Y,M,D,Ho,Mi,Se,_Ml),
		Date = datime(Time,Y,M,D,Ho,Mi,Se),
		( Arch='i386-win32'  ->
			Host = 'i386-win32'
			;
			getenv( 'HOST', Host )
		)
		;
		host_name( Host ),
		datime( Date )
	).

test( Scrambled, Solution ) :-
	count_letter_occurances_in_words( Scrambled, Counts, Sum ),
	alphabet( AlphaBeto ),
	frequencies( Freqs ),
	write( sum(Sum) ), nl,
	write( counts(Counts) ), nl,
	write( alphabet(AlphaBeto) ), nl,
	write( frequencies(Freqs) ), nl,
	% Scrambled = [H,S,Trd|T],
	% nth1( 23, Scrambled, Trtn ),
	% write( looking_for([Trtn]) ), nl,
	% test( [Trtn], Counts, Sum, AlphaBeto, Freqs, Vars ).
	caesar( Scrambled, Counts, Sum, AlphaBeto, Freqs, _Vars, Solution ).

w(Vars) :-
	test( ["edcba"], 
		 [0'e-10, 0'd-6, 0'c-8, 0'b-3, 0'a-13], 
		 40,
		 "abcde",
		 [17/80,10/80,18/80,10/80,25/80],
		 Vars).  % Words = ["abcde"],

x(Vars) :-
	test( ["dba","cae"], 
		 [0'e-10, 0'd-6, 0'c-8, 0'b-3, 0'a-13], 
		 40,
		 "abcde",
		 [17/80,10/80,18/80,10/80,25/80],
		 Vars). % Words = ["edc",bca"]

z(_Codes-Vs) :-
	print_pfd_pr_order( [17/80,10/80,18/80,10/80,25/80], "abcde" ),
	member( Vs, "abc" ).

my_remove_duplicates([], []).
my_remove_duplicates([Head|Tail1], [Head|Tail2]) :- 
	delete(Tail1, Head, Residue),
     my_remove_duplicates(Residue, Tail2).

:- ensure_loaded( '../../src/load_pfd' ).

caesar( EncodedWs, Counts, Sum, AlphaBeto, Freqs, Vars, DecodedWs ) :-
	export_words( scrambled, EncodedWs ),
	write( scrabled_counters(Counts) ), nl,
	write( original_freqs(one_thoushanth_units) ), nl,
	print_pfd_pr_order( Freqs, AlphaBeto ),
	write( '---' ), nl,
	mould_n_pins( Counts, Sum, AlphaBeto, Freqs, Vars, Codes ),
	write( codes(Codes) ), nl,
	write( vars(Vars) ), nl,
	word_codes_to_vars( EncodedWs, Codes, Vars, WordVars ),
	decode_words( WordVars, Codes, Vars, DecodedWs ).

min_cardinality_word( [H|T], Word, UnqVsWord, Rest ) :-
	my_remove_duplicates( H, NoDplH ),
	word_cardinality( NoDplH, 0, HCard ),
	min_cardinality_word( T, NoDplH, H, HCard, WordReps ),
	WordReps = UnqVsWord-Word,
	select( Word, [H|T], Rest ), 
	!.

min_cardinality_word( [], NoDpWrd, Word, _CurrMin, NoDpWrd-Word ).
min_cardinality_word( [H|T], CrNoDpWrd, CurrW, CurrMin, WordReps  ) :-
	my_remove_duplicates( H, NoDplH ),
	word_cardinality( NoDplH, 0, HCard ),
	( HCard =< CurrMin -> 
		NxMin is HCard,
		NxNdpWrd = NoDplH,
		NxW = H
		;
		NxMin is CurrMin,
		NxNdpWrd = CrNoDpWrd,
		NxW = CurrW
	),
	min_cardinality_word( T, NxNdpWrd, NxW, NxMin, WordReps ).

word_exists_in_dictionary(Guess) :- 
	bb_get( unscrambled, Words ), % test
	% unscrambled( Words ), % test
	% Words = ["abcde"],	% w
	% Words =["edc","bca"], 	% x
	( member(Guess,Words) -> 
			% write( yes_word), nl
			true
			;
			fail
			% write( no_word ), nl, fail
	).

mould_n_pins( [], _Sum, _Dom, _Prbs, [], [] ).
mould_n_pins( [Code-Occs|T], Sum, Dom, Prbs, [Hv|Tv], [Code|Tc] ) :-
	% Hv pin proximity(Dom,Prbs,Occs),
	% pin( Hv, proximity(Dom,Prbs,Occs) ),
	Hv pin proximity(Dom,Prbs,Occs/Sum),
	mould_n_pins( T, Sum, Dom, Prbs, Tv, Tc ).

word_cardinality( [], Card, Card ).
word_cardinality( [H|T], AccCard, WCard ) :-
	domain_cardinality( H, HCard ),
	NxAcc is HCard + AccCard,
	word_cardinality( T, NxAcc, WCard ).

decode_words( [], _Codes, _Vars, [] ).
decode_words( [H|T], Codes, Vars, [Guess|Tsolv] ) :-
	min_cardinality_word( [H|T], WordVs, UnqVsWord, Rest ),
	word_vars_to_codes( WordVs, Vars, Codes, WordCs ),
	% write( wordcodes(WordCs) ), nl,
	label_n( pr_order, UnqVsWord, Vals, _Prob, _AccProb ),
	% write( vals(Vals) ), nl,
	new_wordcodes_to_guess( UnqVsWord, WordVs, Vals, Guess ),
	word_exists_in_dictionary( Guess ),
	export_words( current_word, [WordCs]),write( ' - ' ), 
	export_words( proposed_decoding, [Guess]), nl,
	% export_words( intelligent_guess, [Guess]), nl,
	% statistics( walltime, Wstat ),
	% statistics( runtime, Rstat  ),
	% write( stats(wall(Wstat),run(Rstat)) ), nl,
	decode_words( Rest, Codes, Vars, Tsolv ).

/* 
decode_words( [], _Codes, _Vars, [] ).
decode_words( [H|T], Codes, Vars, [Guess|Tsolv] ) :-
	min_cardinality_word( [H|T], WordVs, UnqVsWord, Rest ),

	% wordcodes_to_vars( Word, Codes, Vars, WordVs ),
	word_vars_to_codes( WordVs, Vars, Codes, WordCs ),
	% export_words( current_word, [WordCs]),
	% write( current_word(Word) ), nl,
	% write( varsinword(WordVs) ), nl,
	label_n( pr_order, UnqVsWord, Vals, _Prob, _AccProb ),
	% write( unq_w_vs(UnqVsWord)-w_vs(WordVs)-vals(Vals) ),
	% wordcodes_to_guess( WordCs, Codes, Vars, WordVs, Vals, Guess ),
	new_wordcodes_to_guess( UnqVsWord, WordVs, Vals, Guess ),
	word_exists_in_dictionary( Guess ),
	export_words( current_word, [WordCs]),write( ' - ' ), 
	export_words( intelligent_guess, [Guess]), nl,
	statistics( walltime, Wstat ),
	statistics( runtime, Rstat  ),
	write( stats(wall(Wstat),run(Rstat)) ), nl,
	decode_words( Rest, Codes, Vars, Tsolv ).
*/

% :- asserta( pfd_message('Type \'test.\' to run the caesar demo.') ), nl.
% :- demo([main]),halt.
% :- main_caesar( 1,44 ).
:- (main,halt) ; halt.
